@@ -5,13 +5,15 @@ use Teng::Row;
 use Class::Accessor::Lite
     rw => [ qw(
         tables
+        namespace
     ) ]
 ;
 
 sub new {
     my ($class, %args) = @_;
     bless {
-        tables => {},
+        tables    => {},
+        namespace => '',
         %args,
     }, $class;
 }
@@ -54,6 +56,12 @@ sub camelize {
     join('', map{ ucfirst $_ } split(/(?<=[A-Za-z])_(?=[A-Za-z])|\b/, $s));
 }
 
+sub prepare_from_dbh {
+    my ($self, $dbh) = @_;
+
+    $_->prepare_from_dbh($dbh) for values %{$self->{tables}};
+}
+
 1;
 
 __END__
